<?

/*

Nanoweb FastCGI support module
==============================

Copyright (C) 2002 Vincent Negrier aka. sIX <six@aegis-corp.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

define(FCGI_VERSION_1, 1);

define(FCGI_BEGIN_REQUEST, 1);
define(FCGI_ABORT_REQUEST, 2);
define(FCGI_END_REQUEST, 3);
define(FCGI_PARAMS, 4);
define(FCGI_STDIN, 5);
define(FCGI_STDOUT, 6);
define(FCGI_STDERR, 7);
define(FCGI_DATA, 8);
define(FCGI_GET_VALUES, 9);
define(FCGI_GET_VALUES_RESULT, 10);

class mod_fcgi {

	function mod_fcgi() {

		$this->modtype="parser_FCGI";
		$this->modname="FastCGI support (beta)";
	
	}

	function build_fcgi_packet($type, $content) {

		$clen=strlen($content);
				
		$packet=chr(FCGI_VERSION_1);
		$packet.=chr($type);
		$packet.=chr(0).chr(1); // Request id = 1
		$packet.=chr((int)($clen/256)).chr($clen%256); // Content length
		$packet.=chr(0).chr(0); // No padding and reserved
		$packet.=$content;

//		techo("SEND: type=$type, len=$clen [".addcslashes($packet, "\0..\37!@\177..\377")."]");

		return($packet);
	
	}

	function build_fcgi_nvpair($name, $value) {

		$nlen=strlen($name);
		$vlen=strlen($value);
		
		$nvpair=chr($nlen);

		if ($vlen<128) $nvpair.=chr($vlen);
		else $nvpair.=chr((int)$vlen/16777216).chr((int)($vlen/65536)%256).chr((int)($vlen/256)%256).chr((int)($vlen%256));
		
		$nvpair.=$name.$value;

		return($nvpair);
	
	}

	function decode_fcgi_packet($data) {

		$ret["version"]=(int)$data[0];
		$ret["type"]=(int)$data[1];
		$ret["length"]=((int)$data[4]*256)+(int)$data[5];
		$ret["content"]=substr($data, 8);

//		techo("RECV: [".addcslashes($data, "\0..\37!@\177..\377")."]");

		return($ret);
	
	}

	function parser_open($dummy_content, $args, $filename, &$rq_err, &$cgi_headers, &$add_ctype) {

		global $conf;
		
		// Connect to FastCGI server

		$fcgi_server=explode(":", $args);

		if (!$this->sck=fsockopen($fcgi_server[0], $fcgi_server[1])) {

			$rq_err=500;
			return (false);

		}

		// Begin session
		
		$begin_rq_packet=chr(0).chr(1).chr(0).chr(0).chr(0).chr(0).chr(0).chr(0);
		fwrite($this->sck, $this->build_fcgi_packet(FCGI_BEGIN_REQUEST, $begin_rq_packet));

		// Build params
		
		$fcgi_params_packet.=$this->build_fcgi_nvpair("SERVER_SOFTWARE", SERVER_STRING);
		$fcgi_params_packet.=$this->build_fcgi_nvpair("SERVER_NAME", ($GLOBALS["vhost"]=="global"?$conf["global"]["servername"]:$GLOBALS["vhost"]));
		$fcgi_params_packet.=$this->build_fcgi_nvpair("GATEWAY_INTERFACE", "FastCGI/1.0");
		$fcgi_params_packet.=$this->build_fcgi_nvpair("SERVER_PROTOCOL", HTTP_VERSION);
		$fcgi_params_packet.=$this->build_fcgi_nvpair("SERVER_PORT", $conf["global"]["listenport"]);
		$fcgi_params_packet.=$this->build_fcgi_nvpair("REQUEST_METHOD", $GLOBALS["http_action"]);
		$fcgi_params_packet.=$this->build_fcgi_nvpair("PATH_INFO", basename($filename));
		$fcgi_params_packet.=$this->build_fcgi_nvpair("PATH_TRANSLATED", realpath($filename));
		$fcgi_params_packet.=$this->build_fcgi_nvpair("SCRIPT_FILENAME", realpath($filename));
		$fcgi_params_packet.=$this->build_fcgi_nvpair("SCRIPT_NAME", $GLOBALS["http_uri"]);
		$fcgi_params_packet.=$this->build_fcgi_nvpair("QUERY_STRING", $GLOBALS["query_string"]);
		$fcgi_params_packet.=$this->build_fcgi_nvpair("REMOTE_HOST", $GLOBALS["remote_host"]);
		$fcgi_params_packet.=$this->build_fcgi_nvpair("REMOTE_ADDR", $GLOBALS["remote_ip"]);
		$fcgi_params_packet.=$this->build_fcgi_nvpair("REMOTE_PORT", $GLOBALS["remote_port"]);
		$fcgi_params_packet.=$this->build_fcgi_nvpair("AUTH_TYPE", $GLOBALS["auth_type"]);

		if ($rq_hdrs=$GLOBALS["htreq_headers"]) while (list($key, $val)=each($rq_hdrs)) putenv("HTTP_".str_replace("-", "_", $key)."=".$val);
		
		if ($GLOBALS["http_action"]=="POST" && $GLOBALS["htreq_content"]) {
		
			$fcgi_params_packet.=$this->build_fcgi_nvpair("CONTENT_TYPE", $rq_hdrs["CONTENT-TYPE"]);
			$fcgi_params_packet.=$this->build_fcgi_nvpair("CONTENT_LENGTH", $rq_hdrs["CONTENT-LENGTH"]);

			$stdin_content=$GLOBALS["htreq_content"];
			
		} else $stdin_content="";

		// Send params
		
		fwrite($this->sck, $this->build_fcgi_packet(FCGI_PARAMS, $fcgi_params_packet));
		fwrite($this->sck, $this->build_fcgi_packet(FCGI_PARAMS, ""));
		
		// Build and send stdin flow

		if ($stdin_content) fwrite($this->sck, $this->build_fcgi_packet(FCGI_STDIN, $stdin_content));
		fwrite($this->sck, $this->build_fcgi_packet(FCGI_STDIN, ""));

		// Read answers from fastcgi server

		$content="";
		
		do {

			$packet=fread($this->sck, 8192);
			$resp=$this->decode_fcgi_packet($packet);
			if ($resp["type"]==FCGI_STDOUT) $content.=$resp["content"];
		
		} while ($resp["type"]==FCGI_STDOUT);
		
		if (($p1=strpos($content, "\r\n\r\n"))!==false) {
		
			$headers=explode("\n", trim(substr($content, 0, $p1)));
			$content=substr($content, $p1+4);

		}

		$GLOBALS["out_encoded"]=false;
		$GLOBALS["http_resp"]="";
		
		for($a=0;$a<=count($headers)-1;$a++) if ($s=trim($headers[$a])) {

			if (substr($s, 0, 5)=="HTTP/") {

				$hd_key="";
				$tmp=explode(" ", $s);
				$rq_err=(int)$tmp[1];
			
			} else {
			
				$hd_key=substr($s, 0, strpos($s, ":"));
				$hd_val=trim(substr($s,strpos($s, ":")+1));

			}

			switch(strtoupper($hd_key)) {

				case "CONTENT-TYPE": $add_ctype=$hd_val;
				break;

				case "LOCATION": $rq_err=302;
				break;

				case "SET-COOKIE": $cgi_headers["cookies"][]=$hd_val;
				break;

				case "CONTENT-ENCODING": $GLOBALS["out_encoded"]=true;
				break;

				case "STATUS":
				$GLOBALS["http_resp"]=$hd_val;
				break;

			}
			
			if (($hd_key) && (strtoupper($hd_key)!="SET-COOKIE")) $cgi_headers[$hd_key]=$hd_val;

		}

		$this->parsed_output.=$content;

	}

	function parser_get_output() {

		$tmp=$this->parsed_output;
		$this->parsed_output="";
		return($tmp);
	
	}
	
	function parser_eof() {

		return($this->parsed_output=="");
	
	}
	
	function parser_close() {

		fclose($this->sck);

	}

}

?>
