#!/usr/local/bin/php -q
<?

/*

Nanoweb aka. aEGiS PHP nano web server
======================================

Copyright (C) 2002 Vincent Negrier aka. sIX <six@aegis-corp.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

error_reporting(E_ERROR | E_PARSE);

define(VERSION, "1.4.3");

define(DEFAULT_CONFIG_FILE, "nanoweb.conf");
define(DEFAULT_LISTEN_ADDR, "0.0.0.0");
define(DEFAULT_LISTEN_DEFAULT_PORT, 80);
define(DEFAULT_LISTEN_QUEUE, 20);
define(DEFAULT_MIMETYPES, "/etc/mime.types");
define(DEFAULT_CONTENT_TYPE, "text/html");
define(DEFAULT_REQUEST_TIMEOUT, 15);
define(DEFAULT_AUTH_FILE, ".nwauth");
define(DEFAULT_DOCROOT, "./");
define(DEFAULT_LOGFILE, "./access.log");
define(DEFAULT_FBWELCOMEFILE, ".welcome");

define(SCK_WRITE_PACKET_SIZE, 4096);
define(SCK_READ_PACKET_SIZE, 2048);

define(HTTP_VERSION, "HTTP/1.1");

define(SERVER_STRING, "aEGiS_nanoweb/".VERSION);

$HTTP_HEADERS=array(	200 => "200 OK",
			302 => "302 Found",
			304 => "304 Not Modified",
			400 => "400 Bad Request",
			401 => "401 Unauthorized",
			403 => "403 Forbidden",
			404 => "404 Not Found",
			500 => "500 Internal Server Error",
			501 => "501 Not Implemented");

$TEST_FUNCS=array(	"pcntl_fork"	=> true, 
			"socket_create"	=> true, 
			"posix_setuid"	=> true, 
			"gzencode"	=> false);

$conf_defaults=array(	"listeninterface"	=> DEFAULT_LISTEN_ADDR,
			"listenport"		=> DEFAULT_LISTEN_PORT,
			"listenqueue"		=> DEFAULT_LISTEN_QUEUE,
			"mimetypes"		=> DEFAULT_MIMETYPES,
			"requesttimeout"	=> DEFAULT_REQUEST_TIMEOUT,
			"authfile"		=> DEFAULT_AUTH_FILE,
			"documentroot"		=> DEFAULT_DOCROOT,
			"log"			=> DEFAULT_LOGFILE,
			"fbwelcomefile" 	=> DEFAULT_FBWELCOMEFILE);

$conf_vhosts_propagate=array(	"documentroot",
			"directoryindex",
			"authfile",
			"serveradmin",
			"user",
			"group",
			"log",
			"filebrowser",
			"fbshowdotfiles",
			"fbwelcomefile");

// Functions

function _parseconfig_close_part(&$conf, $cpart) {

	global $conf_defaults, $conf_vhosts_propagate;
	
	reset($conf_defaults);
	reset($conf_vhosts_propagate);
	
	if ($cpart=="global") while (list($key, $dval)=each($conf_defaults)) if (!$conf[$cpart][$key]) $conf[$cpart][$key]=$dval;
	while (list($key, $pkey)=each($conf_vhosts_propagate)) if (!$conf[$cpart][$pkey]) $conf[$cpart][$pkey]=$conf["global"][$pkey];
	
	return(true);

}

function parseconfig($conf_arr) {

	if ($conf_arr) while (list($key, $str)=each($conf_arr)) {

		switch ($str[0]) {

			case "#":
			case ";":
			case "\n":
			break;
			
			case "[":
			if ($cpart) _parseconfig_close_part($conf, $cpart); 
			$cpart=substr(trim($str), 1, -1);
			break;

			default:

			$cnfk=trim(strtolower(substr($str, 0, strpos($str, "="))));
			$cnfv=trim(substr($str, strpos($str,"=")+1, strlen($str)));
			
			switch ($cnfk) {

				case "documentroot": $conf[$cpart][$cnfk]=realpath($cnfv);
				break;

				case "loadmodule": $conf["loadmodules"][]=$cnfv;
				break;
				
				default: $conf[$cpart][$cnfk]=$cnfv;

			}
			
			break;
		
		}
		
	}

	_parseconfig_close_part($conf, $cpart);
	
	return($conf);

}

function load_modules($conf) {

	while (list($key, $modname)=each($conf["loadmodules"])) {

		@include_once($modname);
		$clsname=basename($modname, ".php");
		$modloaded=class_exists($clsname);

		if ($modloaded) {
		
			$tmp=&new $clsname;
			$modules[$tmp->modtype][]=&$tmp;
			techo("loaded module : ".$tmp->modname);
			$lm++;

		} else {

			techo("WARN: unable to load module '".$clsname."'");
		
		}
	
	}

	return($modules);

}

function techo($s) {

	echo date("Ymd-His")." - $s\n";
	flush();

}

function errexit($s) {

	techo("FATAL: ".$s);
	exit;

}

function check_auth($authname, &$br_name) {

	$authfile=file($authname);
	$authdata["realm"]=trim($authfile[0]);
	$br_name=$authdata["realm"];

	for ($a=1;$a<=count($authfile);$a++) {
		
		$lp=explode(":", trim($authfile[$a]));
		$authdata["users"][$lp[0]]["login"]=$lp[0];
		$authdata["users"][$lp[0]]["des-password"]=$lp[1];

	}

	$hash=$authdata["users"][$GLOBALS["auth_user"]]["des-password"];
	$thash=crypt($GLOBALS["auth_pass"] ,substr($hash, 0, 2));

	if ($hash==$thash) return($GLOBALS["auth_user"]); else return(false);

}

function loadfile($filename, $extension, &$rq_err, &$cgi_headers, &$add_ctype, $bypass_auth=false, $return_str_only=false) {
	
	global $conf, $docroot, $modules;
	
	$add_ctype=false;
	
	if (strpos($filename, $conf[$GLOBALS["vhost"]]["authfile"])!==false) $rq_err=403; else {
	
		if (@is_readable($s_authname=$docroot.$conf[$GLOBALS["vhost"]]["authfile"])) $authname=$s_authname;
		else $authname=dirname($filename)."/".$conf[$GLOBALS["vhost"]]["authfile"];

		$GLOBALS["logged_user"]="";
		
		if (@is_readable($authname) && !$bypass_auth) {

			// HTTP Auth needed
			
			if ($auth=check_auth($authname, $realm_name)) {

				// Authorized

				$GLOBALS["logged_user"]=$auth;

			} else {

				// Unauthorized

				$GLOBALS["logged_user"]=" ";
				$rq_err=401;
				$cgi_headers["WWW-Authenticate"]="Basic realm=\"".$realm_name."\"";

			}
		
		}
		
		// Test If-Modified-Since request header (client has a cached copy)
		
		$process_rq=true;
		
		if ($lmdate=$htreq_headers["IF-MODIFIED-SINCE"] && !$conf["global"]["defineext".strtolower($rq_file["extension"])]) {

			$lmdate=strtotime($lmdate);
			$fdate=filemtime($docroot.$http_uri);

			if ($fdate<=$lmdate) {

				$rq_err=304;
				$process_rq=false;
			
			}
		
		}

		if ($process_rq && $GLOBALS["logged_user"]!=" ") {
		
			chdir(dirname(realpath($filename)));
			$filename=basename($filename);
			
			$f=fopen($filename, "r");
			$content=fread($f, filesize($filename));
			fclose($f);

			if (($parser=trim($conf["global"]["defineext".strtolower($extension)]))) {

				// Parsed content
				
				if (strpos($parser, " ")!==false) {
				
					$ps_type=substr($parser, 0, strpos($parser, " "));
					$ps_arg=substr($parser, strpos($parser, " ")+1);
					$ps_arg=str_replace("$"."FILENAME", $filename, $ps_arg);

				} else $ps_type=$parser;

				if (is_object($ps=&$modules["parser_".$ps_type][0])) $content=$ps->parser_open($content, $ps_arg, $filename, $rq_err, $cgi_headers, $add_ctype);
			
			} else {
	
				// Static content

				$cgi_headers["Last-Modified"]=gmdate("D, d M Y H:i:s T", filemtime($filename));
			
			}
		
		}
	
	}
	
	if (is_object($ps) && !$return_str_only) return($ps); else return($content);

}

function nanoweb_shutdown($sig_no=SIGTERM) {

	global $sock, $mypid;
	
	if ($sock) socket_close($sock);
	if (!$mypid) techo("daemon stopped\n");
	exit;

}

function reload_config($sig_no=SIGHUP) {

	global $conf, $conffile, $mypid, $modules;

	if (!$mypid) {
	
		$conf=parseconfig(file($conffile));
		techo("configuration file (".$conffile.") reloaded");
		$modules=load_modules($conf);

	}

}

function send_response($response, &$sck_connected) {

	global $msgsock, $sckv2;
	
	if ($sckv2) $fdset=array($msgsock); 
	else $fdset=$GLOBALS["fdset"];
	
	$sent_len=0;
	$resp_len=strlen($response);

	while($sent_len<$resp_len && $sck_connected) {

		if (($sent_len+SCK_WRITE_PACKET_SIZE)>$resp_len) $size=$resp_len-$sent_len; else $size=SCK_WRITE_PACKET_SIZE;
		if (($ret=@socket_write($msgsock, substr($response, $sent_len, $size), $size))>0) $sent_len+=$ret;
		if (!socket_select($read=NULL, $fdset, $except=NULL, 10)) $sck_connected=false;

	}

	return($sent_len);

}

// Begin

set_time_limit(0);

techo("aEGiS nanoweb/".VERSION." (c)2002 by sIX <six@aegis-corp.org>");

if (version_compare(phpversion(), "4.2.0")>=0) $sckv2=true;

while (list($f_name, $f_mandatory)=each($TEST_FUNCS)) if (!is_callable($f_name)) {

	if ($f_mandatory) errexit("function '".$f_name."' not available, aborting");
	else techo("WARN: function '".$f_name."' not available");

}

if ($argv[1]) $conffile=$argv[1]; else $conffile=DEFAULT_CONFIG_FILE;
if (!is_readable($conffile)) errexit("unable to read configuration (".$conffile."), aborting");
$conf=parseconfig(file($conffile));
$modules=load_modules($conf);

if ((!$conf["global"]["disablegzip"]) && (!is_callable("gzencode"))) {
	
	techo("WARN: gzencode() unavailable, turning off gzip encoding");
	$conf["global"]["disablegzip"]=true;

}

pcntl_signal(SIGTERM, "nanoweb_shutdown");
pcntl_signal(SIGHUP, "reload_config");

if (!@is_readable($conf["global"]["mimetypes"])) {

	techo("WARN: unable to read mime types file (".$conf["global"]["mimetypes"]."), using internals");

	$mime["default"]="text/plain";
	$mime["html"]="text/html";
	$mime["gif"]="image/gif";
	$mime["jpeg"]="image/jpeg";
	$mime["png"]="image/png";
	$mime["tgz"]="application/gtar";

} else if ($mimetypes=file($conf["global"]["mimetypes"])) {
	
	while (list($key, $s)=each($mimetypes)) if (trim($s) && ($s[0]!="#")) if (ereg("([a-zA-Z0-9/.-]+)[ \t]+([a-zA-Z0-9 -]+)", $s, $res)) if ($exts=explode(" ", trim($res[2]))) while (list($key2, $ext)=each($exts)) if (trim($res[1]) && trim($ext)) $mime[$ext]=trim($res[1]);

}

if (($sock = @socket_create(AF_INET, SOCK_STREAM, 0))<0) errexit("socket() failed: ".socket_strerror($sock));
if (($ret = @socket_bind($sock, $conf["global"]["listeninterface"], $conf["global"]["listenport"]))<0) errexit("bind() failed: ".socket_strerror($ret));
if (($ret = @socket_listen($sock, $conf["global"]["listenqueue"]))<0) errexit("listen() failed: ".socket_strerror($ret));

socket_set_nonblock($sock);

techo("running and accepting connections on port ".$conf["global"]["listenport"]);

while (true) {

	$cnx=false;

	while (!$cnx) {

		pcntl_waitpid(-1, $cstat, WNOHANG);
		if (is_resource($msgsock=@socket_accept($sock))) $cnx=true; else usleep(1000);
	
	}
	
	$pid=pcntl_fork();

	if ($pid===0) {
	
		$mypid=posix_getpid();
		
		set_time_limit($conf["global"]["childlifetime"]);
		
		socket_set_nonblock($msgsock);
		@socket_getpeername($msgsock, $remote_ip, $remote_port);

		if (!$sckv2) {
		
			$fdset=socket_fd_alloc();
			socket_fd_set($fdset, $msgsock);

		}

		$remote_host=$remote_ip;

		if ($conf["global"]["hostnamelookups"]) if ($rhost=@gethostbyaddr($remote_ip)) $remote_host=$rhost;

		while ($cnx) {
		
			$sck_connected=true;
			$rq_finished=false;
			$lcount=0;
			$wstart=(int)date("U");
			$buf="";
			
			while (!$rq_finished && $sck_connected) {
			
				$tmp=@socket_read($msgsock, SCK_READ_PACKET_SIZE);
				
				if ($tmp || $tmp==="0") {
					
					$wstart=(int)date("U");
					$buf.=$tmp;
					$pn=0;

					if ($dp=strpos($buf, "\r\n\r\n")) $pn=4;
					else if ($dp=strpos($buf, "\n\n")) $pn=2;
						
					if ($pn) $rq_finished=true;

				} else if (socket_last_error($msgsock)!=11) {
					
					$sck_connected=false;

				} else if ($lcount++==50) {

					$lcount=0;
					$r_elapsed=(int)date("U")-$wstart;

					if ($r_elapsed>$conf["global"]["requesttimeout"]) {

						$sck_connected=false;
					
					}
				
				} else usleep(1000);

			}

			if ($sck_connected) {

				if (strlen($buf)!=$dp+4) $add_req=substr($buf, $dp+4); else $add_req="";
				$tmp_arr=explode("\n",substr($buf, 0, $dp));

				$htreq=explode(" ", trim($tmp_arr[0]));
				$http_action=strtoupper($htreq[0]);
				$http_protocol=explode("/", $htreq[2]);
				$http_version=$http_protocol[1];

				// Decode HTTP request headers
				
				unset($htreq_headers);
				
				for($a=1;$a<=count($tmp_arr);$a++) if ($s=trim($tmp_arr[$a])) {

					$hd_key=trim(strtoupper(substr($s, 0, strpos($s, ":"))));
					$hd_val=trim(substr($s,strpos($s, ":")+1));

					$htreq_headers[$hd_key]=$hd_val;

				}

				if ($http_action=="POST" && $htreq_headers["CONTENT-LENGTH"]) {

					// Read request content if POST
					
					$rq_finished=false;
					$lcount=0;
					$wstart=(int)date("U");
					$buf=$add_req;
					if (strlen($buf)>=$htreq_headers["CONTENT-LENGTH"]) $rq_finished=true;

					while (!$rq_finished && $sck_connected) {

						$tmp=@socket_read($msgsock, SCK_READ_PACKET_SIZE);
						
						if ($tmp || $tmp==="0") {
							
							$wstart=(int)date("U");
							$buf.=$tmp;
							if (strlen($buf)>=$htreq_headers["CONTENT-LENGTH"]) $rq_finished=true;

						} else if (socket_last_error($msgsock)!=11) {
							
							$sck_connected=false;

						} else if ($lcount++==50) {

							$lcount=0;
							$r_elapsed=(int)date("U")-$wstart;

							if ($r_elapsed>$conf["global"]["requesttimeout"]) {

								$sck_connected=false;
							
							}
						
						} else usleep(1000);

						if ($tmp==="0") $sck_connected=true;
					
					}

					$htreq_content=substr($buf, 0, $htreq_headers["CONTENT-LENGTH"]);
				
				}
				
				if ($sck_connected) {

					if ($auth_hdr=$htreq_headers["AUTHORIZATION"]) {

						// Decode HTTP Authentication header

						$dtmp=explode(" ", $auth_hdr);
						$auth_type=$dtmp[0];
						$auth_lp=explode(":", base64_decode($dtmp[1]));
						$auth_user=$auth_lp[0];
						$auth_pass=$auth_lp[1];
					
					} else $auth_type=$auth_user=$auth_pass="";
					
					// Test for Keep-Alive header
					
					$keepalive=(strtolower(trim($htreq_headers["CONNECTION"]))=="keep-alive" && (int)$conf["global"]["keepalive"]>1);
					if ($keepalive && ++$rq_count>=(int)$conf["global"]["keepalive"]) $keepalive=false;
					$cnx=$keepalive;
					
					// Test for Host header
					
					$host=trim(strtolower($htreq_headers["HOST"]));
					if (($p=strpos($host, ":"))!==false) $host=substr($host, 0, $p);
					if (is_array($conf[$host])) $vhost=$host; else $vhost="global";

					if ($setgid=@posix_getgrnam($conf[$vhost]["group"])) posix_setgid($setgid["gid"]);
					if ($setuid=@posix_getpwnam($conf[$vhost]["user"])) posix_setuid($setuid["uid"]);

					$complete_path=$http_uri=$rq_file=$out_headers="";
					$rq_err=$out_content_length=0;
					
					$docroot=$conf[$vhost]["documentroot"];
					if ($docroot[strlen($docroot)-1]!="/") $docroot.="/";
					
					$http_protocol=$htreq[count($htreq)-1];
					
					if ($p1=$htreq[1]) {

						$p1=explode("?", $p1);
					
						$complete_path=$http_uri=ltrim(urldecode($p1[0]), "/");
						$query_string=$p1[1];
						if (realpath($docroot.$http_uri) && strpos(realpath($docroot.$http_uri), realpath($docroot))===false) $http_uri="";

						if (is_dir($docroot.$http_uri) && $conf[$vhost]["directoryindex"]) {
							
							$dis=explode(" ", $conf[$vhost]["directoryindex"]);
							$dfound=false;
							
							while ((list($key, $diname)=each($dis)) && !$dfound) if (is_readable($docroot.$http_uri."/".$diname)) {
									
								$http_uri.="/".$diname;
								$dfound=true;

							}

						}

						$rq_file=pathinfo($http_uri);

					}

					$out_contenttype=DEFAULT_CONTENT_TYPE;
					$out_content="";
					unset($out_add_headers);

					switch ($http_action) {

						case "GET":
						case "POST":
						case "HEAD":

						if (is_dir($docroot.$http_uri)) {

							// Directory browsing

							if ($http_uri[strlen($http_uri)-1]!="/") $http_uri.="/";

							if (@is_readable($s_authname=$docroot.$conf[$GLOBALS["vhost"]]["authfile"])) $authname=$s_authname;
							else $authname=$docroot.$http_uri.$conf[$GLOBALS["vhost"]]["authfile"];

							if (@is_readable($authname)) $auth=check_auth($authname, $br_realm); else $auth=true;

							if ($conf[$vhost]["filebrowser"] && $auth) {

								$rq_err=200;
								$lf="<html><head><title>Index of ".$http_uri."</title></head><body bgcolor=\"#ffffff\" text=\"#000000\"><table><tr><td bgcolor=\"#ffffff\" class=\"title\"><font size=\"4\" face=\"Verdana, Arial, sans-serif\"><b>Index of ".$http_uri."</b></font><br><br><br></td></tr></table><table cellspacing=\"2\" cellpadding=\"2\" width=\"615\"><tr><td><b>Name</b></td><td><b>Last modified</b></td><td><b>Size</b></td></tr><tr><td colspan=\"3\"><hr noshade align=\"left\"></td></tr>";

								if (@is_readable($wfn=$docroot.$http_uri.$conf[$vhost]["fbwelcomefile"])) {

									$wfc=implode("<br>", file($wfn));
									$lf.="<br><font size=\"3\">".$wfc."</font><br><br>";
								
								}
								
								$hnd=opendir(realpath($docroot.$http_uri));

								while ($f=readdir($hnd)) {

									$fi=stat($docroot.$http_uri.$f);
									$isdir=is_dir($docroot.$http_uri.$f);

									if ($f=="..") $dname="Parent directory";
									else $dname=$f;

									if ((($f[0]!="." || $f=="..") || $conf[$vhost]["fbshowdotfiles"]) && ($f!=".") && !($f==".." && $http_uri=="/")) $lf.="<tr><td><font face=\"Verdana, Arial\" size=\"2\"><a href=\"/".($http_uri!="/"?$http_uri:"").urlencode($f)."\">".$dname."</a></font></td><td><font face=\"Verdana, Arial\" size=\"2\">".date("d-M-Y H:i:s", $fi[9])."</font></td><td align=\"right\"><font face=\"Verdana, Arial\" size=\"2\">".($isdir?"&lt;&nbsp;dir&nbsp;&gt;":$fi[7]." bytes")."</font></td></tr>";

								}

								closedir($hnd);
								
								$lf.="<tr><td colspan=\"3\"><hr noshade align=\"left\"><font size=\"1\">".SERVER_STRING."</font></td></tr></table>";
							
							} else if ($auth===false) {
								
								$rq_err=401;
								$out_add_headers["WWW-Authenticate"]="Basic realm=\"".$br_realm."\"";
							
							} else $rq_err=404;
						
						} else if (!is_file($docroot.$http_uri)) $rq_err=404;
						else if (!is_readable($docroot.$http_uri)) $rq_err=403;
						else {

							$rq_err=200;
							$lf=loadfile($docroot.$http_uri, $rq_file["extension"], $rq_err, $out_add_headers, $add_ctype);
							if ($mimetype=$mime[strtolower($rq_file["extension"])]) $out_contenttype=$mimetype; else $out_contenttype=DEFAULT_CONTENT_TYPE;

						}

						if ($http_action=="HEAD") $lf="";

						break;

						case "OPTIONS":
						$rq_err=200;
						$out_add_headers["Allow"]="GET, ".($conf["global"]["defineext".strtolower($rq_file["extension"])]?"POST, ":"")."HEAD, OPTIONS";
						break;

						case "PUT":
						case "DELETE":
						case "TRACE": $rq_err=501;
						break;

						default: $rq_err=400;
						break;

					}

					if ($rq_err>=400) {
						
						if (($errordoc=trim($conf["global"]["errordocument".$rq_err])) && (@is_readable($docroot.$errordoc))) {

								$errext=substr(strrchr($errordoc, "."), 1);
								$lf=loadfile($docroot.$errordoc, $errext, $rq_err, $out_add_headers, $add_ctype, true);

								if ($mimetype=$mime[strtolower($errext)]) $out_contenttype=$mimetype; else $out_contenttype=DEFAULT_CONTENT_TYPE;
							
						} else {
						
							$out_contenttype="text/html";
							$lf="<html><head><title>".$HTTP_HEADERS[$rq_err]."</title></head><body><h1>".$HTTP_HEADERS[$rq_err]."</h1>";
							if ($http_uri) $lf.="An error occured trying to access resource <i>".htmlentities($http_uri)."</i>.<br><br>";
							if ($admmail=$conf[$vhost]["serveradmin"]) $lf.="Please contact the server <a href=\"mailto:".$admmail."\">administrator</a> if the problem is abnormal and persists.<br><br>";
							$lf.="<hr><font size=\"1\">".SERVER_STRING."</font></body></html>";

						}
					
						$cnx=false;
					
					} else if ($rq_err>=300) {

						$lf=false;
					
					}
					
					$out_headers=HTTP_VERSION." ".($http_resp?trim($http_resp):$HTTP_HEADERS[$rq_err])."\r\n";
					$out_headers.="Date: ".date("D, d M Y H:i:s T")."\r\n";
					$out_headers.="Server: ".SERVER_STRING." (PHP/".phpversion().")\r\n";
					if (!$add_ctype && $lf!==false) $out_headers.="Content-Type: ".$out_contenttype."\r\n";

					if ($out_add_headers["cookies"]) while(list($key, $val)=each($out_add_headers["cookies"])) $out_headers.="Set-Cookie: ".$val."\r\n";

					if ($out_add_headers) {
						
						while(list($key, $val)=each($out_add_headers)) if ($key!="cookies") $out_headers.=$key.": ".$val."\r\n";
						unset($out_add_headers);
					
					}

					if (is_string($lf) && !$out_encoded && strpos($htreq_headers["ACCEPT-ENCODING"], "gzip")!==false && !$conf["global"]["disablegzip"]) {
						
						$gz_content=gzencode($lf);

						if (strlen($gz_content)<(strlen($lf)*0.95)) {

							// Send compressed content only if it's smaller than 95% of original
							
							$out_headers.="Content-Encoding: gzip\r\n";
							$lf=$gz_content;

						}
					
					}

					if (is_string($lf)) $out_content_length=strlen($lf);
					
					if ($keepalive) {
					
						$chunked=($http_version=="1.1" && is_object($lf));
						
						$out_headers.="Connection: Keep-Alive\r\n";
						$out_headers.="Keep-Alive: timeout=".(int)$conf["global"]["requesttimeout"].", max=".(int)($conf["global"]["keepalive"]-$rq_count)."\r\n";
						if ($chunked) $out_headers.="Transfer-Encoding: chunked\r\n";
						
						if ($out_content!==false && !$chunked && !is_object($lf)) $out_headers.="Content-Length: ".$out_content_length."\r\n";

					} else {
						
						$chunked=false;
						$out_headers.="Connection: close\r\n";

					}
					
					if (is_object($lf)) {

						// Dynamic content stream

						$out_content="";
						$sent_len=0;

						if ($chunked) send_response($out_headers."\r\n", $sck_connected);
						
						while (!$lf->parser_eof() && $sck_connected) if ($buf=$lf->parser_get_output()) {
							
							$out_content.=$buf;

							if ($chunked) {

								$chunk_header=dechex(strlen($buf))."\r\n";
								$metasize=strlen($chunk_header+2);
								$rbytes=send_response($chunk_header.$buf."\r\n", $sck_connected);
								$sent_len=$sent_len+($rbytes-$metasize);

							}

						}

						$lf->parser_close();

						if ($chunked) {
							
							$meta_len=0;
							send_response("0\r\n\r\n", $sck_connected);
							

						} else {

							$out_headers.="Content-Length: ".strlen($out_content)."\r\n";
							$response=$out_headers."\r\n".$out_content;
							$meta_len=strlen($out_headers)+2;

							$sent_len=send_response($response, $sck_connected);
						
						}

					} else {
					
						// Static content
						
						$out_content=$lf;
						$response=$out_headers."\r\n".$out_content;
						$meta_len=strlen($out_headers)+2;

						$sent_len=send_response($response, $sck_connected);
					
					}

					if (!$sck_connected) $cnx=false;
					if (($sent_content_length=$sent_len-$meta_len)<0) $sent_content_length=0;

					if ($nb_loggers=count($modules["log"])) {

						for ($a=0;$a<$nb_loggers;$a++) $modules["log"][$a]->log_hit($vhost, $remote_host, $remote_ip, $logged_user, trim($tmp_arr[0]), $rq_err, $sent_content_length, $htreq_headers["REFERER"], $htreq_headers["USER-AGENT"]);

					}
					
				}
			
			} else {

				$cnx=false;
			
			}

		}
		
		if (!$sckv2) {
		
			socket_fd_clear($fdset, $msgsock);
			socket_fd_free($fdset);

		}

		socket_close($msgsock);
		exit;

	} else if ($pid==-1) techo("ERROR: unable to pcntl_fork()");
	
}

?>
